% :- type file        ==  atom.

% :- type wrspec      ---> wr(min_overlap, max_overlap, ring_length, list(word)).
% :- type min_overlap == int.
% :- type max_overlap == int.
% :- type ring_length == int.
% :- type word        == atom.
% :- type overlap     == int.
% :- type wopair      ---> word-overlap.
% :- type ring        == list(wopair).

% wordring(WRSPEC, RING):
% RING a WRSPEC specifikacionak megfelelo szogyuru
% :- pred wordring(wrspec::in, ring::out).

:- module(wordring_keret,
	  [
	   helyes_feladvany/1,  % helyes_feladvany(wrspec::in)
	   wordring_in/2,	% wordring_in(file::in, wrspec::out)
	   wordring_out/3,	% wordring_out(file::in, ring::in,
	                        %         ring_length::in)
	   megold/2,		% megold(file::in, file::in)
	   stopper/2		% stopper(file::in, file::in)
	  ]
	 ).

:- use_module(library(lists), [member/2,append/3]).

%-------------- Bejaratok ------------------------------------

% wordring_in(+File, -Feladvany):
% A File-bol beolvashato a Feladvany altal megadott feladat.
wordring_in(File, Feladvany) :-
	open_file(File, read, Stream),
	call_cleanup(read_in(Stream, Characters),
		     close(Stream)),
	parse_data(Characters, Feladvany).

% wordring_out(+File, +Szogyuru, +Hossz):
% A Hossz hosszusagu Szogyuru megoldasat kiirja a File allomanyba.
wordring_out(File, Szogyuru, _Hossz) :-
	open_file(File, write, Stream),
	current_output(OOut),
	set_output(Stream),
	call_cleanup(pretty_print(Szogyuru),
		     (set_output(OOut),close(Stream))).

% megold(+FileIn, +FileOut):
% A FileIn allomanybol beolvasott feladat minden megoldasat
% kiirja a FileOut allomanyba.
megold(FileIn, FileOut) :-
	wordring_in(FileIn, Feladvany),
	ensure_loaded(user:wordring),
	open_file(FileOut, write, Stream),
	current_output(OOut),
	set_output(Stream),
	call_cleanup(print_solutions(Feladvany),
		     (set_output(OOut),close(Stream))).

% stopper(+FileIn, +FileOut):
% Mint megold/2, de a futashoz szukseges idot is kiirja.
stopper(FileIn, FileOut) :-
	ensure_loaded(user:wordring),
	statistics(runtime, [T0|_]),
	megold(FileIn, FileOut),
	statistics(runtime, [T1|_]),
	T is T1 - T0,
	format('Feladat: ~a, futasi ido ~3d s~n', [FileIn,T]).

helyes_feladvany(wr(Min,Max,Len,Words)) :-
	% Az alapveto tipusok helyesek:
	integer(Min),
	integer(Max),
	integer(Len),
	atom_lista(Words),
	% Nehany relacio igaz:
	Min =< Max,
	Len >= Min.	

%-------------- Ellenorzesek ---------------------------------

helyes_pont(S-O, MaxS-MaxO) :-
	integer(S), S >=1, S =< MaxS,
	integer(O), O >=1, O =< MaxO.

atom_lista([]).
atom_lista([A|As]) :-
	atom(A),
	atom_lista(As).

%-------------- Allomanyok kezelese --------------------------

% open_file(+File, +Mode, -Stream):
% A File allomanyt megnyitja a Stream folyam formajaban Mode modban.
open_file(user, M, S) :-
	!, standard_stream(M, S).
open_file(File, Mode, Stream) :-
	open(File, Mode, Stream).

% standard_stream(?Mode, ?Stream):
% A Mode modu szabvanyos folyam a Stream.
standard_stream(read, user_input).
standard_stream(write, user_output).

%-------------- Beolvasas ------------------------------------

% read_in(+S, -L):
% S folyambol beolvashato karakterek listaja L.
read_in(S, L) :-
	get_code(S, C),
	read_in(C, S, L).

% read_in(+C, +S, -L):
% Ha C az S folyambol elozoleg beolvasott karakter, akkor L az S-bol
% beolvashato maradek karakterek listaja, C-t is beleertve.
% Ha C = -1, akkor L = [].
read_in(C, _, L) :-
	C =:= -1, !, L = [].
read_in(C, S, [C|L]) :-
	get_code(S, C1),
	read_in(C1, S, L).

% parse_data(+Chars, -Feladvany):
% Chars fuzerbol kielemezheto egy Feladvany feladvanyleiro
parse_data(Chars, Feladvany) :-
	descriptor(Feladvany, Chars, Rest),
	correct_syntax(Rest),
	helyes_feladvany(Feladvany), !.
parse_data(_, _) :-
	warning('A bemeneti allomany nem megfelelo formatumu.',[]),
	fail.

% correct_syntax(+Rest):
% A Rest fuzer a bemenet helyes maradeka, azaz ures.
correct_syntax(Rest) :-
	Rest = [_|_], !,
	chars_of_type(mid_line, L, Rest, _),
	warning('Nem vart karakter(ek): "~s"', [L]),
	fail.
correct_syntax(_).

%-------------- DCG elemzes ----------------------------------

% Kielemezheto egy feladvany-leiro
descriptor(wr(Min,Max,Len,Words)) -->
	nat_number(Min), nat_number(Max), nat_number(Len),
	newline,
	word_list(Words),
	chars_of_type(layout, _).

% Kielemezheto egy (tizes szamrendszerbeli) Int termeszetes szam, amit
% esetleg szokozok eloznek meg.
nat_number(Int) -->
	chars_of_type(horizontal, _),
	chars_of_type(digit, Ds),
	{Ds = [_|_], number_codes(Int, Ds)}.

% Kielemezheto egy (kis)betukbol allo szo, amit
% esetleg szokozok eloznek meg.
alpha_word(Word) -->
	chars_of_type(horizontal, _),
	chars_of_type(alpha, Ls),
	{Ls = [_|_], atom_codes(Word, Ls)}.

% word_list(Words)
% Kielemezheto egy atomokbol allo Words lista.
word_list([W|Ws]) -->
	alpha_word(W),
	newline, !,
	word_list(Ws).
word_list([]) --> [].

newline -->
	chars_of_type(horizontal, _),
	char_of_type(vertical, _).


% layout_char(Dir, C):
% horizontal_layout_char(C): C  egy vizszintes iranyu
% nem lathato karakter.
horizontal_layout_char(0' ).
horizontal_layout_char(0'\t).
horizontal_layout_char(0'\r).

% vertical_layout_char(C): C  egy fuggoleges iranyu
% nem lathato karakter.
vertical_layout_char(0'\n).
vertical_layout_char(0'\f).
vertical_layout_char(0'\v).

% Kielemezheto egy Type tipusu C karakter.
char_of_type(Type, C) -->
	[C], {char_type(Type, C)}.

% chars_of_type(Type, Cs) -->
% Kielemezheto Type tipusu karakterek egy Cs listaja.	
chars_of_type(Type, [C|Cs]) -->
	char_of_type(Type, C), !,
	chars_of_type(Type, Cs).
chars_of_type(_, []) --> [].

% char_type(Type, C) :
% A C karakter Type tipusu.
char_type(layout, C) :-
	(   vertical_layout_char(C) -> true
	;   horizontal_layout_char(C)
	).
char_type(digit, D) :-
	D >= 0'0, D =< 0'9.
% Csak kisbetuket fogadunk el.
char_type(alpha, A) :-
	A >= 0'a, A =< 0'z.
char_type(mid_line, C) :-
	\+ vertical_layout_char(C).
char_type(horizontal, C) :-
	horizontal_layout_char(C).
char_type(vertical, C) :-
	vertical_layout_char(C).

%-------------- Kiiratas -------------------------------------

% print_solutions(+Feladvany):
% A Tabla altal megadott feladat osszes megoldasat
% kiirja az aktualis kimenetre.
print_solutions(Feladvany) :-
	user:wordring(Feladvany, Szogyuru),
	pretty_print(Szogyuru),
	fail.
print_solutions(_).

% pretty_print(Feladvany, Szogyuru)
% Kiirja a Feladvany feladvany Szogyuru altal megadott megoldasait
pretty_print(Szogyuru) :-
	print_wordlist(Szogyuru, 0).

print_wordlist([], T) :-
	tab(T), write('|...'), nl.
print_wordlist([W-N|WNs], T) :-
	tab(T), write(W), nl,
	atom_length(W, L),
	T1 is T + L - N,
	print_wordlist(WNs, T1).

%-------------- Seged-eljarasok ------------------------------

% warning(+Txt, +Args):
% Kiirja a Txt szoveggel es Args argumentumlistaval
% megadott figyelmezteto szoveget.
warning(Txt, Args) :-
	print_message(warning, format(Txt, Args)).

%-------------- ISO kompatibilitas ---------------------------

% Az alabbi cel az ISO Prolog szabvanynak meg nem megfelelo, de
% DEC10 kompatiblis Prolog rendszerekben valo futtatashoz szukseges.

:- (   predicate_property(get_code(_,_), _) -> true
   ;   assert((get_code(S, C) :- get0(S, C))),
       assert((put_code(C) :- put(C))),
       assert((number_codes(N, C) :- number_chars(N, C)))
   ).
